"""
In principle, one could automate the process for merging data annotations
between binding sites that overlap and need to be merged together. For now, this
function and file does nothing, and merging is handled in other ways.
"""


def generate_merge_func(_):
    """
    Returns the identity function
    :param _: the inputt paramter is ignored

    """
    return lambda x: x
